/* Pango
 * m17n-fc.c: Generic shaper using the m17n library for FreeType-based backends

   Copyright (C) 2004, 2005, 2006
     National Institute of Advanced Industrial Science and Technology (AIST)
     Registration Number H16PRO276

   This file is part of the pango-m17n library.

   The pango-m17n library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The pango-m17n library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the implied warranty
   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the pango-m17n library; if not, write to the
   Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA.  */

#include <string.h>
#include <stdlib.h>

#include <pango/pango-engine.h>
#include <pango/pango-utils.h>
#include <pango/pangofc-font.h>

#include <m17n-gui.h>
#include <m17n-misc.h>

#ifdef G_LOG_DOMAIN
#undef G_LOG_DOMAIN
#endif
#define G_LOG_DOMAIN "Pango-M17N-FC"

/* Control printing of debug information.  It is initialized from the
   environment variable PANGO_M17N_FC_DEBUG.  Currently, any positive
   value turns on all debug messages.  */
static int debug_level;

/* Commonly used frame on nulldevice.  */
static MFrame *frame;

/* Commonly used "generic" fontset.  */
static MFontset *fontset;

/* Initialize this module.  The main task is to open a frame on
   nulldevice.  */

static void
init_module ()
{
  const char *envvar = g_getenv ("PANGO_M17N_FC_DEBUG");
  MPlist *param;
  MFace *face;

  debug_level = envvar ? atoi (envvar) : 0;

  param = mplist ();
  face = mface ();
  fontset = (mdatabase_find (Mfontset, msymbol ("generic"), Mnil, Mnil)
	     ? mfontset ("generic") : mfontset (NULL));
  mface_put_prop (face, Mfontset, fontset);
  mplist_put (param, Mface, face);
  mplist_put (param, Mdevice, Mnil);
  frame = mframe (param);
  m17n_object_unref (face);
  m17n_object_unref (param);
}

/* Finalize this module by freeing frame.  */

static void
fini_module ()
{
  m17n_object_unref (frame);
  m17n_object_unref (fontset);
}

/* No extra fields needed for these structures.  */
typedef PangoEngineShape      M17NEngineFc;
typedef PangoEngineShapeClass M17NEngineFcClass;

#define SCRIPT_ENGINE_NAME "M17NScriptEngineFc"
#define RENDER_TYPE PANGO_RENDER_TYPE_FC

static PangoEngineScriptInfo m17n_scripts[] = {
  { PANGO_SCRIPT_ARABIC,   "*" },
  { PANGO_SCRIPT_ARMENIAN, "*" },
  { PANGO_SCRIPT_BENGALI,  "*" },
  { PANGO_SCRIPT_BOPOMOFO, "*" },
  { PANGO_SCRIPT_CHEROKEE, "*" },
  { PANGO_SCRIPT_COPTIC,   "*" },
  { PANGO_SCRIPT_CYRILLIC, "*" },
  { PANGO_SCRIPT_DESERET,  "*" },
  { PANGO_SCRIPT_DEVANAGARI, "*" },
  { PANGO_SCRIPT_ETHIOPIC, "*" },
  { PANGO_SCRIPT_GEORGIAN, "*" },
  { PANGO_SCRIPT_GOTHIC,   "*" },
  { PANGO_SCRIPT_GREEK,    "*" },
  { PANGO_SCRIPT_GUJARATI, "*" },
  { PANGO_SCRIPT_GURMUKHI, "*" },
  { PANGO_SCRIPT_HAN,      "*" },
  { PANGO_SCRIPT_HANGUL,   "*" },
  { PANGO_SCRIPT_HEBREW,   "*" },
  { PANGO_SCRIPT_HIRAGANA, "*" },
  { PANGO_SCRIPT_KANNADA,  "*" },
  { PANGO_SCRIPT_KATAKANA, "*" },
  { PANGO_SCRIPT_KHMER,    "*" },
  { PANGO_SCRIPT_LAO,      "*" },
  { PANGO_SCRIPT_MALAYALAM, "*" },
  { PANGO_SCRIPT_MYANMAR,  "*" },
  { PANGO_SCRIPT_OGHAM,    "*" },
  { PANGO_SCRIPT_OLD_ITALIC, "*" },
  { PANGO_SCRIPT_ORIYA,     "*" },
  { PANGO_SCRIPT_RUNIC,     "*" },
  { PANGO_SCRIPT_SINHALA,   "*" },
  { PANGO_SCRIPT_SYRIAC,    "*" },
  { PANGO_SCRIPT_TAMIL,     "*" },
  { PANGO_SCRIPT_TELUGU,    "*" },
  { PANGO_SCRIPT_THAANA,    "*" },
  { PANGO_SCRIPT_THAI,      "*" },
  { PANGO_SCRIPT_TIBETAN,   "*" },
  { PANGO_SCRIPT_CANADIAN_ABORIGINAL, "*" },
  { PANGO_SCRIPT_YI,       "*" },

#ifdef PANGO_SCRIPT_BRAILLE
  { PANGO_SCRIPT_BRAILLE,  "*" },
  { PANGO_SCRIPT_CYPRIOT,  "*" },
  { PANGO_SCRIPT_LIMBU,    "*" },
  { PANGO_SCRIPT_OSMANYA,  "*" },
  { PANGO_SCRIPT_SHAVIAN,  "*" },
  { PANGO_SCRIPT_LINEAR_B, "*" },
  { PANGO_SCRIPT_TAI_LE,   "*" },
  { PANGO_SCRIPT_UGARITIC, "*" },
#endif	/* PANGO_SCRIPT_BRAILLE */

#ifdef PANGO_SCRIPT_NEW_TAI_LUE
  { PANGO_SCRIPT_NEW_TAI_LUE, "*" },
  { PANGO_SCRIPT_BUGINESE, "*" },
  { PANGO_SCRIPT_GLAGOLITIC, "*" },
  { PANGO_SCRIPT_TIFINAGH, "*" },
  { PANGO_SCRIPT_SYLOTI_NAGRI, "*" },
  { PANGO_SCRIPT_OLD_PERSIAN, "*" },
  { PANGO_SCRIPT_KHAROSHTHI, "*" },
#endif	/* PANGO_SCRIPT_NEW_TAI_LUE */

  { PANGO_SCRIPT_UNKNOWN,  "*" },

  { PANGO_SCRIPT_COMMON,   "" }
};

static PangoEngineInfo script_engines[] = {
  {
    SCRIPT_ENGINE_NAME,
    PANGO_ENGINE_TYPE_SHAPE,
    RENDER_TYPE,
    m17n_scripts, G_N_ELEMENTS (m17n_scripts)
  }
};

static char *
m17n_fc_get_family (PangoFont *pango_font)
{
  PangoFcFont *pango_fc_font = PANGO_FC_FONT (pango_font);
  FcChar8 *family;

  if (FcPatternGetString (pango_fc_font->font_pattern, FC_FAMILY, 0, &family)
      == FcResultMatch)
    return (char *) family;
  return "";
}

/* Convert PANGO_LANGUAGE to m17n-lib's language symbol.  If m17n-lib
   doesn't know about PANGO_LANGUAGE, return Mnil.  */

static MSymbol
m17n_fc_get_language (PangoLanguage *pango_language)
{
  MSymbol language;

  if (pango_language)
    {
      const char *lang = pango_language_to_string (pango_language);

      language = msymbol (lang);
      if (! msymbol_get (language, Mlanguage))
	{
	  if (strlen (lang) <= 2)
	    language = Mnil;
	  else
	    {
	      /* Remove region part (e.g. "zh_CN" -> "zh").  */
	      char shortlang[3];
	  
	      shortlang[0] = lang[0], shortlang[1] = lang[1], shortlang[2] = 0;
	      language = msymbol (shortlang);
	      if (! msymbol_get (language, Mlanguage))
		language = Mnil;
	    }
	}
    }
  else
    language = Mnil;
  return language;
}


/* Return a list of FONT-SPECs specified in the current fontset for
   displaying SCRIPT in LANGUAGE.  Callers must unref the returned
   plist.  */

MPlist *
m17n_fc_lookup_fontset (MSymbol script, MSymbol language)
{
  MPlist *plist = mfontset_lookup (fontset, script, language, Mnil);

  if (mplist_key (plist) == Mnil)
    {
      /* List up languages that have entries for SCRIPT.  */
      MPlist *ll = mfontset_lookup (fontset, script, Mt, Mnil);  

      if (! ll)
	{
	  /* No entry for SCRIPT.  Return fallback FONT-SPECs. */
	  m17n_object_unref (plist);
	  plist = mfontset_lookup (fontset, Mnil, Mnil, Mnil);
	}
      else
	{
	  if (mplist_length (ll) == 1)
	    {
	      /* If there's just one language, return FONT-SPECs for it.  */
	      m17n_object_unref (plist);
	      language = mplist_key (ll);
	      if (language == Mt)
		language = Mnil;
	      plist = mfontset_lookup (fontset, script, mplist_key (ll), Mnil);
	    }
	  else
	    /* Sumup FONT-SPECs for all languages.  */
	    for (; mplist_key (ll) != Mnil; ll = mplist_next (ll))
	      {
		MPlist *pl;

		language = mplist_key (ll);
		if (language == Mt)
		  language = Mnil;
		for (pl = mfontset_lookup (fontset, script, language, Mnil);
		     mplist_key (pl) != Mnil; pl = mplist_next (pl))
		  mplist_add (plist, mplist_key (pl), mplist_value (pl));
	    }
	}
    }
  return plist;
}


typedef struct
{
  PangoFcFont *pango_fc_font;
  MFont *font;
} M17NFontInfo;

static void
m17n_fc_free_font_info (gpointer data)
{
  M17NFontInfo *font_info = data;

  if (font_info->font)
    {
      mfont_close (font_info->font);
      pango_fc_font_unlock_face (font_info->pango_fc_font);
    }
  g_free (data);
}

static MFont *
m17n_fc_get_font (PangoFont *pango_font)
{
  GQuark info_id = g_quark_from_string ("m17n-font-info");
  M17NFontInfo *font_info;

  font_info = g_object_get_qdata (G_OBJECT (pango_font), info_id);
  if (! font_info)
    {
      PangoFcFont *pango_fc_font = PANGO_FC_FONT (pango_font);
      FT_Face ft_face = pango_fc_font_lock_face (pango_fc_font);

      font_info = g_new (M17NFontInfo, 1);
      font_info->pango_fc_font = pango_fc_font;
      font_info->font = mfont_encapsulate (frame, Mfreetype, ft_face);
      g_object_set_qdata_full (G_OBJECT (pango_font), info_id, font_info,
			       (GDestroyNotify) m17n_fc_free_font_info);
    }
  return font_info->font;
}


/* Shaper function.  */

static void 
m17n_fc_engine_shape (PangoEngineShape *engine,
		      PangoFont        *pango_font,
		      const char       *text,
		      int              length,
		      const PangoAnalysis    *analysis,
		      PangoGlyphString *glyphs)
{
  /* Symbols for character property `category'.  */
  static MSymbol MZs, MCf;
  static MDrawControl control;
  MSymbol language;
  PangoGlyphInfo *g;
  MText *mt;
  MDrawGlyph *m_glyphs;
  int nchars, nglyphs;
  MFont *font;
  int i;
  const char *p;
  int *offsets;

  g_return_if_fail (pango_font != NULL);
  g_return_if_fail (text != NULL);
  g_return_if_fail (length >= 0);
  g_return_if_fail (analysis != NULL);

  if (debug_level > 0)
    {
      char *msg = alloca (length + 1);

      if (msg)
	{
	  memcpy (msg, text, length);
	  msg[length] = '\0';
	}
      else
	msg = "...";
      g_debug ("shape \"%s\"", msg);
    }

  if (! MZs)
    {
      MZs = msymbol ("Zs");
      MCf = msymbol ("Cf");
    }

  language = m17n_fc_get_language (analysis->language);

  mt = mtext_from_data (text, length, MTEXT_FORMAT_UTF_8);
  nchars = mtext_len (mt);
  offsets = alloca (sizeof (int) * nchars);
  g_return_if_fail (offsets != NULL);
  for (i = 0, p = text; i < nchars; i++, p = g_utf8_next_char (p))
    offsets[i] = p - text;

  font = m17n_fc_get_font (pango_font);
  g_return_if_fail (font != NULL);

  mtext_put_prop (mt, 0, nchars, Mfont, font);
  if (language != Mnil)
    mtext_put_prop (mt, 0, nchars, Mlanguage, language);

  control.enable_bidi = 1;
  m_glyphs = alloca (sizeof (MDrawGlyph) * nchars * 2);
  g_return_if_fail (m_glyphs != NULL);
  if (mdraw_glyph_list (frame, mt, 0, nchars, &control, m_glyphs,
			nchars * 2, &nglyphs) < 0)
    {
      m_glyphs = alloca (sizeof (MDrawGlyph) * nglyphs);
      g_return_if_fail (m_glyphs != NULL);
      mdraw_glyph_list (frame, mt, 0, nchars, &control, m_glyphs,
			nglyphs, &nglyphs);
    }

  pango_glyph_string_set_size (glyphs, nglyphs);

  for (i = 0, g = glyphs->glyphs; i < nglyphs; i++, g++)
    {
      if (m_glyphs[i].glyph_code >= 0)
	{
	  g->glyph = m_glyphs[i].glyph_code;
	  g->geometry.x_offset = m_glyphs[i].x_off * PANGO_SCALE;
	  g->geometry.y_offset = m_glyphs[i].y_off * PANGO_SCALE;
	  g->geometry.width = m_glyphs[i].x_advance * PANGO_SCALE;
	}
      else
	{
	  int c = mtext_ref_char (mt, m_glyphs[i].from);
	  MSymbol category = mchar_get_prop (c, Mcategory);
	  PangoRectangle logical_rect;

	  if (category == MZs)
	    {
	      g->glyph = 0;
	      g->geometry.width = m_glyphs[i].x_advance * PANGO_SCALE;
	    }
	  else if (category == MCf)
	    {
	      g->glyph = 0;
	      g->geometry.width = 0;
	    }
	  else
	    {
	      g->glyph = (pango_fc_font_get_unknown_glyph
			  (PANGO_FC_FONT (pango_font), c));
	      pango_font_get_glyph_extents (pango_font, g->glyph, NULL,
					    &logical_rect);
	      g->geometry.width = logical_rect.width;
	    }
	  g->geometry.x_offset = 0;
	  g->geometry.y_offset = 0;
	}
      g->attr.is_cluster_start
	= (i == 0
	   || m_glyphs[i - 1].from != m_glyphs[i].from);
      glyphs->log_clusters[i] = offsets[m_glyphs[i].from];
    }

  m17n_object_unref (mt);
}

/* Check if PANGO_FONT is suitable for the character WC.  */

static PangoCoverageLevel
m17n_fc_engine_covers (PangoEngineShape *engine,
		       PangoFont        *pango_font,
		       PangoLanguage    *pango_language,
		       gunichar          wc)
{
  static gunichar last_wc = 0;
  /* Symbols for character property `script'.  */
  static MSymbol Mlatin, Mcommon, Minherited;
  PangoCoverage *coverage;
  PangoCoverageLevel result;
  MSymbol script, language;
  MPlist *spec_list, *p;
  MFont *font;

  if (! Mlatin)
    {
      Mlatin = msymbol ("latin");
      Mcommon = msymbol ("common");
      Minherited = msymbol ("inherited");
    }

  if (debug_level > 0 && wc != last_wc)
    {
      g_debug ("covers for U+%04X", wc);
      last_wc = wc;
    }

  coverage = pango_font_get_coverage (pango_font, pango_language);
  result = pango_coverage_get (coverage, wc);
  pango_coverage_unref (coverage);

  if (result == PANGO_COVERAGE_NONE)
    {
      if (debug_level > 0)
	g_debug ("  %s none", m17n_fc_get_family (pango_font));
      return result;
    }

  language = m17n_fc_get_language (pango_language);
  if (wc < 0x100)
    script = Mlatin;
  else
    {
      script = mchar_get_prop ((int) wc, Mscript);
      if (script == Mcommon || script == Minherited)
	script = mchar_get_prop ((int) wc, msymbol ("block"));
    }

  font = m17n_fc_get_font (pango_font);
  spec_list = m17n_fc_lookup_fontset (script, language);

  result = PANGO_COVERAGE_NONE;
  for (p = spec_list; mplist_key (p) != Mnil; p = mplist_next (p))
    if (mfont_match_p (font, mplist_value (p)))
      {
	result = PANGO_COVERAGE_EXACT;
	break;
      }
  m17n_object_unref (spec_list);
  if (debug_level > 0)
    g_debug ("  (%s,%s): %s %s",
	     (script ? msymbol_name (script) : ""),
	     (language ? msymbol_name (language) : ""),
	     m17n_fc_get_family (pango_font),
	     result == PANGO_COVERAGE_EXACT ? "exact" : "none");

  return result;
}

static void
m17n_engine_fc_class_init (PangoEngineShapeClass *class)
{
  class->script_shape = m17n_fc_engine_shape;
  class->covers = m17n_fc_engine_covers;
}

PANGO_ENGINE_SHAPE_DEFINE_TYPE (M17NEngineFc, m17n_engine_fc,
				m17n_engine_fc_class_init, NULL)

void 
PANGO_MODULE_ENTRY(init) (GTypeModule *module)
{
  m17n_engine_fc_register_type (module);

  M17N_INIT ();
  init_module ();
}

void 
PANGO_MODULE_ENTRY(exit) (void)
{
  fini_module ();
  M17N_FINI ();
}

void 
PANGO_MODULE_ENTRY(list) (PangoEngineInfo **engines,
			  int              *n_engines)
{
  *engines = script_engines;
  *n_engines = G_N_ELEMENTS (script_engines);
}

PangoEngine *
PANGO_MODULE_ENTRY(create) (const char *id)
{
  return (!strcmp (id, SCRIPT_ENGINE_NAME)
	  ? g_object_new (m17n_engine_fc_type, NULL)
	  : NULL);
}
